import suiteninit, suitenout
from suitenamedefs import Suite, Residue, Bin, Cluster, Issue, failMessages
from suiteninit import args, bins, MAX_CLUSTERS
from suiteninit import normalWidths, satelliteWidths
from suiteninput import readResidues, readKinemageFile, buildSuites
from suitenutil import hyperEllipsoidDistance

import sys, os
import numpy as np
from math import cos, pi

#                           suitename.py
# ***************************************************************
# NOTICE: This is free software and the source code is freely
# available. You are free to redistribute or modify under the
# conditions that (1) this notice is not removed or modified
# in any way and (2) any modified versions of the program are
# also available for free.
#               ** Absolutely no Warranty **
# Copyright (C) 2007 David C. Richardson
# ***************************************************************

# 0.2.070524 preserve chi-1 and chi, so could preserve eta, theta
# 0.3.070525 general read dangle record for, e.g.,  eta, theta
# 0.3.070628 triage reports zeta-1, epsilon-1, delta-1,... Ltriage codes
# 0.3.070803 notes: rearranged suitenhead.h/janesviews ...
#                  put something in to say what veiws mean
#  what are masters e and d ????
# 0.3.070919 3g wannabe (tRNA TpseudoUC loop)
# 0.3.110606 range of delta updated by S.J.
# 01/07/2014 S.J. updated so that it can take input with alternate conformations, *nd by default will calculate the suite for altA
# 09/18/2014 S.J. updated so that suitename will ignore DNA residues


version = "suitename.0.6.012521"
dbCounter = 0
dbTarget = 10000  # triggers extra output on this suite

# A collection of variables used for output
class OutNote:
    pass

outNote = OutNote()
outNote.version = version
outNote.comment = ""
outNote.wannabes = 0
outNote.outliers = 0


# ***main()******************************************************************
def main():
    global dbCounter  # for debugging KPB 210222

    # 1. read the input
    if args.infile != "":
        inFile = open(args.infile)
#    elif sys.gettrace() is not None: # how to detect debugger present
    else:
        inFile = sys.stdin

    if args.suitein or args.suitesin:
        suites = readKinemageFile(inFile)
        if len(suites) == 0:
            sys.stderr.write("read no suites: perhaps wrong type of kinemage file\n")
            sys.exit(1)
    else:
        residues = readResidues(inFile)
        if len(residues) == 0:
            sys.stderr.write("read no residues: perhaps wrong alternate code\n")
            sys.exit(1)
        suites = buildSuites(residues)
        suites = suites[:-1]

    # 2. process the suites
    for s in suites:
        if not s.validate():
            if args.test:
                sys.stderr.write(f"! failed validation: {s.pointID}\n")
            suitenout.write1Suite(
                s, bins[13], bins[13].cluster[0], 0, 0, " tangled ", "", "", "", ""
            )
            continue

        # At this point we have a complete suite
        bin, issue, text, pointMaster = evaluateSuite(s)
        pointColor = "white"
        if bin is None:
            s.cluster = bins[0].cluster[0]
            bins[0].cluster[0].count += 1
            suitenout.write1Suite(
                s, bins[0], bins[0].cluster[0], 0, 0, text, issue, "", "", ""
            )
        else:
            memberPack = membership(bin, s)
            (cluster, distance, suiteness, notes, comment,
                pointMaster, pointColor) = memberPack
            suitenout.write1Suite(
                s, bin, cluster, distance, suiteness, notes, issue, comment, 
                pointMaster, pointColor)
            s.suiteness = suiteness
            s.distance = distance
            s.notes = notes
        s.pointMaster = pointMaster
        s.pointColor = pointColor
        dbCounter += 1

    finalStats()
    suitenout.writeFinalOutput(suites, outNote)


# *** evaluateSuite and its tools ***************************************

# Boundaries of various angle ranges
epsilonmin = 155
epsilonmax = 310  # 070130
delta3min = 60
delta3max = 105  #  changed by S.J. on 06/06/2011
delta2min = 125
delta2max = 165
gammapmin = 20
gammapmax = 95  # max 070326
gammatmin = 140
gammatmax = 215  # max 070326
gammammin = 260
gammammax = 335  # max 070326
alphamin = 25
alphamax = 335
betamin = 50
betamax = 290
zetamin = 25
zetamax = 335

# triage table for yes-no angles:
# each of these filters, applied to a suite, will provide a
# true or false answer as to whether this angle is in a reasonable range.
# pointMaster is for grouping points in kinemage display

# data per line: angle index, min, max, code, text, pointMaster
triageFilters = {
    "epsilon": (2, epsilonmin, epsilonmax, Issue.EPSILON_M, "e out", "E"),
    "alpha":   (4, alphamin, alphamax, Issue.ALPHA, " a out", "T"),
    "beta":    (5, betamin, betamax, Issue.BETA, " b out", "T"),
    "zeta":    (3, zetamin, zetamax, Issue.ZETA_M, " z out", "T"),
}

def triage(selector, suite):
    # if angle lies outside the acceptable range, triage immediately
    filter = triageFilters[selector]
    index, min, max, failCode, failText, pointMaster = filter
    if suite.angle[index] < min or suite.angle[index] > max:
        return False, failCode, failText, pointMaster
    else:
        return True, None, None, ""


# The more complex angles are handled by a "sieve".
# A sieve will determine whether an angle is within one of several ranges
# and provide an appropriate code indicating the range.
# This is handled by the sift() function.
sieveDelta = (
    (delta3min, delta3max, 3),
    (delta2min, delta2max, 2),
)

sieveGamma = (
    (gammatmin, gammatmax, "t"),
    (gammapmin, gammapmax, "p"),
    (gammammin, gammammax, "m"),
)

def sift(sieve, angle, failCode):
    for filter in sieve:
        min, max, code = filter
        if min <= angle <= max:
            return code, "", ""
    failMessage = failMessages[failCode]
    return None, failCode, failMessage


def evaluateSuite(suite):
    global bins

    # The order of triage operations, though it may seem arbitrary,
    # was carefully chosen by the scientists.
    ok, failCode, notes, pointMaster = triage("epsilon", suite)
    if not ok:
        return None, failCode, notes, pointMaster

    # Angles with several meaningful ranges:
    # for each angle, find out which range it lies in, or none
    # this becomes a selector to help choose a bin
    puckerdm, failCode, notes = sift(sieveDelta, suite.deltaMinus, Issue.DELTA_M)
    if not puckerdm:
        return None, failCode, notes, "D"

    puckerd, failCode, notes = sift(sieveDelta, suite.delta, Issue.DELTA)
    if not puckerd:
        return None, failCode, notes, "D"

    gammaname, failCode, notes = sift(sieveGamma, suite.gamma, Issue.GAMMA)
    if not gammaname:
        return None, failCode, notes, "T"

    ok, failCode, notes, pointMaster = triage("alpha", suite)
    if not ok:
        return None, failCode, notes, pointMaster

    ok, failCode, notes, pointMaster = triage("beta", suite)
    if not ok:
        return None, failCode, notes, pointMaster

    ok, failCode, notes, pointMaster = triage("zeta", suite)
    if not ok:
        return None, failCode, notes, pointMaster

    # We have pass the test: now use this information to select a bin
    bin = bins[(puckerdm, puckerd, gammaname)]
    # bins is an associated dictionary indexed by the triplet of three angle classifiers
    # each unique triplet of classifiers selects one unique bin, for a total of 12 bins.
    return bin, None, None, ""


# ***membership()***************************************************************

# cluster membership:
# given the bin, we are looking for the correct cluster
def membership(bin, suite):
    matches = np.full(MAX_CLUSTERS, 999.9)
    matchCount = 0
    comment = ""
    pointMaster = ""
    pointColor = "white"

    lDominant = bin.dominant > 0
    if lDominant:  # this bin has a dominant cluster, note it
        dominantJ = bin.dominant
        domCluster = bin.cluster[bin.dominant]

    # find the closest cluster
    # search every cluster in the bin except cluster 0, which is for outliers
    closestD = 999
    for j, c in enumerate(bin.cluster[1:], 1):
        if c.status == "wannabe" and args.nowannabe:
            continue
        distance = hyperEllipsoidDistance(
            suite.angle, bin.cluster[j].angle, 4, normalWidths
        )
        if distance < closestD:
            closestD = distance
            closestJ = j
            closestCluster = c
        matches[j] = distance
        if distance < 1:  # suite could be a member of this cluster
            matchCount += 1

    if matchCount == 1:
        theCluster = closestCluster
        situation = "1-only-one"

    elif matchCount > 1 and not lDominant:
        # dominant cluster is not a possible cluster
        # just output than minimum distance match
        theCluster = closestCluster
        situation = "{matchCount}-None-dom"

    elif matchCount > 1:  # and lDominant
        # find the closest cluster that is not the dominant cluster
        closestNonD = 999
        for j, c in enumerate(bin.cluster[1:], 1):
            if c.status == "wannabe" and args.nowannabe:
                continue
            if matches[j] < closestNonD and c.dominance != "dom":
                closestNonD = matches[j]
                closestJ = j
                theCluster = c

        if theCluster.dominance == "sat":
            # We need to distinguish carefully whether our suite
            # is in the dominant or satellite cluster
            theCluster, closestJ, situation = domSatDistinction(
                suite, domCluster, theCluster, matches, matchCount
            )
        else:
            if matches[dominantJ] < matches[closestJ]:
                closestJ = dominantJ
                theCluster = domCluster
            situation = f"{matchCount}-not-sat"
    else:
        # no match, it's an outlier
        closestJ = 0
        theCluster = closestCluster
        situation = f"outlier distance {closestD:.3}"
        outNote.outliers += 1
        pointMaster = "O"
        pointColor = "white"

    # final computation of suiteness
    # this time we use all 7 dimensions
    if dbCounter >= dbTarget and dbCounter <= dbTarget + 1:  # KPB debug
        print(suite.pointID)
        print(suite.angle)
        print(theCluster.name)
      
    distance = hyperEllipsoidDistance(suite.angle, theCluster.angle, 7, normalWidths)
    # this calculation can assign or deassign a cluster
    if distance <= 1:
        suiteness = (cos(pi * distance) + 1) / 2
        if suiteness < 0.01:
            suiteness = 0.01
    else:
        if closestJ != 0:
            # 7D distance forces this suite to be an outlier
            # so we deassign it here
            closestJ = 0
            comment = f"7D dist {theCluster.name}"
            if theCluster.status == "wannabe":
                comment += " wannabe"
        theCluster = bin.cluster[0]  # outlier
        suiteness = 0

    theCluster.count += 1
    suite.cluster = theCluster
    if theCluster.status == "wannabe" and not args.nowannabe:
        outNote.wannabes = 1  # once set, stays set
    pointColor = 0  # will be handled later!!
    if args.test:
        print(" [suite: %s %s 4Ddist== %f, 7Ddist== %f, suiteness==%f] \n" % \
                (theCluster.name, suite.pointID[:11], closestD, distance, suiteness))
    return theCluster, distance, suiteness, situation, comment, pointMaster, pointColor


def domSatDistinction(suite, domCluster, satCluster, matches, matchCount):
    # if dotproducts both positive, then inbetween
    #      p                  p
    #  dom/___sat  and   dom___\sat
    closestCluster = satCluster
    closestJ = satCluster.ordinal
    dominantJ = domCluster.ordinal

    # use vector properties of numpy.array to determine difference vectors
    domToPoint = domCluster.angle - suite.angle
    satToPoint = satCluster.angle - suite.angle
    domToSat = domCluster.angle - satCluster.angle
    satToDom = -domToSat

    dps = narrowDotProduct(domToPoint, domToSat, 4)
    spd = narrowDotProduct(satToPoint, satToDom, 4)

    if dps > 0 and spd > 0:
        # the trickiest case: point is between dom and sat
        domWidths = normalWidths.copy()
        if args.satellites:
            satWidths = satelliteWidths.copy()
        else:
            satWidths = normalWidths.copy()
        if satCluster.satelliteInfo is not None:
            modifyWidths(domWidths, satWidths, satCluster.satelliteInfo)
        disttodom = hyperEllipsoidDistance(suite.angle, domCluster.angle, 4, domWidths)
        disttosat = hyperEllipsoidDistance(suite.angle, satCluster.angle, 4, satWidths)
        if disttodom < disttosat:
            closestJ = dominantJ
            closestCluster = domCluster
        situation = f"{matchCount}-BETWEEN-dom-sat({disttodom:7.3}|{disttosat:7.3})"
        # else the satellite cluster remains the chosen cluster

    else:
        # the point is not in between
        # just assign by closest standard distance evaluation
        if matches[dominantJ] < matches[closestJ]:
            closestJ = dominantJ
            closestCluster = domCluster
        situation = f"{matchCount}-OUTSIDE-dom-sat"

    return closestCluster, closestJ, situation


# *** Gathering some statistics *********************************************

def finalStats():
    for bin in bins.values():
        for c in bin.cluster:
            bin.count += c.count


# *** The fancy math ********************************************************

# This variable was experimental but we have settled on 3:
power = 3

def hyperEllipsoidDistance(suiteAngles, clusterAngles, nAngles, widthArray):
    global dbCounter
    if nAngles == 4:
        workRange = range(2, 6)
    else:
        workRange = range(1, 8)

    summation = 0
    for k in workRange:
        delta = abs(suiteAngles[k] - clusterAngles[k])
        delta = delta / widthArray[k]
        delToPower = pow(delta, power)
        summation = summation + delToPower
        if dbCounter >= dbTarget and nAngles > 4:  # KPB debug 120221
            sys.stderr.write("db=%3d, k=%d, del=%8.4f, delpower=%10.6f, dpower=%10.6f\n" % 
                        (dbCounter, k, delta, delToPower, summation) )
    result = pow(summation, 1 / power)
    if dbCounter == dbTarget and nAngles > 4:
        sys.stderr.write("final = %7.3f\n" % result)
    return result


# The narrow dot product involves only a subset of the dimensions,
# either 4 or 6. In practice, only 4 is in use.
def narrowDotProduct(a, b, nAngles):
    if nAngles == 4:
        return np.dot(a[2:6], b[2:6])
    else:
        return np.dot(a[1:8], b[1:8])


def modifyWidths(dom, sat, satInfo):
    for m in range(9):
        if satInfo.satelliteWidths[m] > 0:
            sat[m] = satInfo.satelliteWidths[m]
        if satInfo.dominantWidths[m] > 0:
            dom[m] = satInfo.dominantWidths[m]


def showHelpText():
    sys.stderr.write(
        f"""
Version {version}
suitename -flags <stdin >stdout
  or
suitename inputfile -flags >stdout
output flags: [ -report || -string || -kinemage ]
default:  -report

input flags: [ -residuein || -suitein  ]
flags: [ -residuein [ -pointIDfields # ] ] default#=={args.pointidfields}
 OR 
flags: [ -suitein [ -anglefields # ] ]   default#=={args.anglefields}
defaults: -residuein  -pointIDfields {args.pointidfields}

The -residuein format:
label:model:chain:number:ins:type:alpha:beta:gamma:delta:epsilon:zeta
if the file has alternate conformations, then use both -pointIDfields 
    and -altIDfield # to specify the number of pointID fields and which field
    contains the altID
use -altIDval <altID> to specify which alternate conformation to calculate 
    suite for. By default calculated for alt A

-suitein takes a kinemage format,  and uses records from 
    @ballists and/or @dotlists in this format:
{{ptID}} [chi] deltam epsilon zeta alpha beta gamma delta [chi] 
    @dimension in the file, if present, overrides -anglefields

Note dangle trick to make theta,...,eta suites directly

flag: -report [ -chart ]
 suites in order of input, suiteness summary at end
( -chart : NO summary at end, for MolProbity multichart)

flag: -string [-nosequence] [-oneline] 
 3 character per suite string in order of input
    20 per line, ptID of n*20th at end of line
  flag: -nosequence
    only suite names, no Base sequence character
  flag: -oneline
    string all one line, no point IDs

flag: -kinemage
 kinemage of clusters grouped by pucker,pucker ... 
 group {{delta,delta}},subgroup {{gamma}},list {{cluster name}}
  flag: -etatheta or -thetaeta
    kinemage labels theta,eta instead of chi-1,chi

flag: -satellite
  use special general case satellite widths
flag: -nowannabe   
  never assign suites to wannabe clusters
Note: any DNA residues found in the input will be ignored.
""")

main()



from suitenamedefs import Bin, Cluster, Issue, reasons
from suiteninit import args, bins

import sys
from enum import Enum
import numpy as np
from numpy import array

outFile = sys.stdout  # future: might provide a way to change this

reportCountAll = 0
trigCountAll = 0
suitenessSumAll = 0
binSuiteCountAll = 0


def write1Suite(suite, bin, cluster, distance, suiteness, notes, 
                issue, comment, pointMaster, pointColor):
    if args.string:
        string1Suite(suite, cluster)
    elif args.kinemage:
        kinemage1Suite(suite, bin, cluster, notes, distance, suiteness, 
                       issue, comment, pointMaster, pointColor,
        )
    else:
        report1Suite(suite, bin, cluster, notes, distance, suiteness, issue, comment)


def writeFinalOutput(suites, outNote):
    if not args.satellites:
        outNote.comment = " all general case widths, power = 3.00"
    else:
        outNote.comment = " special general case satellite widths, power = 3.00"
    if args.string:
        pass
    elif args.kinemage:
        kinemageFinal(suites, outNote)
    else:
        reportFinal(outNote)


def string1Suite(suite, cluster):
    suiteCount += 1
    if args.nosequence:
        basestring = ":"
    else:
        basestring = suite.base
    outFile.write(f"{cluster.name}{basestring}")


def report1Suite(
    suite, bin, cluster, notes, distance, suiteness, issue, comment
):  # ? LComment, Ltriage
    global reportCountAll, trigCountAll, suitenessSumAll, binSuiteCountAll

    # 1. write one line of output for this suite
    reason = ""
    if issue:
        reason = " " + reasons[issue]
    elif comment:
        reason = " " + comment
        comment = ""
    if cluster.status == "wannabe":
        comment = " wannabe"
    outIDs = ":".join(suite.pointID)
    output = (
        f"{outIDs} {bin.name} {cluster.name} {float(suiteness):5.3f}{reason}{comment}\n"
    )
    outFile.write(output)

    # 2. gather statistics
    reportCountAll += 1

    if bin.ordinal == 0:
        trigCountAll += 1
    elif bin.ordinal < 13:
        suitenessSumAll += suiteness
        binSuiteCountAll += 1

    if cluster.ordinal == 0:
        cluster.suitenessCounts[11] += 1
    else:
        cluster.suitenessSum += suiteness
        # report in statistical baskets at intervals of 0.1:
        # everything from 0 to 0.1 goes in bucket 1
        # ... everything from 0.9 to 1.o goes into bucket 10
        if suiteness == 0:
            bucket = 0
        else:
            bucket = 1 + int(suiteness * 10)
        cluster.suitenessCounts[bucket] += 1


def reportFinal(outNote):
    if not args.chart:
#@deleteLine:
        outFile.write(outNote.comment + "\n")
        suitenessAverage(0)
        if bins[1].cluster[1].count > 0:  # Aform 1a    070325
            suitenessAverage(1)
            suitenessAverage(2)


def suitenessAverage(mode):
    """
    Gather statistics on suiteness.
    12 buckets:
      one for suiteness=0
      one each for divisions by tenths from 0 to 1
      one for outliers
    """
    bucket = np.zeros(12, dtype=int)
    sum = 0
    average = 0
    allCount = 0
    excludedCluster = None

    if mode == 1:
        # cluster 1a all by itself
        comment = " A form (1a)"
        cluster = bins[1].cluster[1]
        sum = cluster.suitenessSum
        for k in range(12):
            bucket[k] = cluster.suitenessCounts[k]
    else:
        if mode == 0:
            # all complete suites
            startWith = 0
            comment = "For all"
            outFile.write(
                "Found {} complete suites derived from {} entries\n".format(
                    binSuiteCountAll + trigCountAll, reportCountAll
                )
            )
            outFile.write(
                "{} suites were triaged, leaving {} assigned to bins\n".format(
                    trigCountAll, binSuiteCountAll
                )
            )
        elif mode == 2:
            # all complete suites except cluster 1a
            startWith = 1
            comment = " non-1a  has"
            excludedCluster = bins[1].cluster[1]
        for i in range(1, 13):
            bin = bins[i]
            # the final bin 13 is for the pseudo-suites with incomplete
            # angles and is ignored
            for cluster in bin.cluster[startWith:]:
                # cluster 0 in every bin is for outliers and is ignored
                if cluster is excludedCluster:
                    continue  # mode 2: ignore cluster 1a
                sum += cluster.suitenessSum
                # print(
                #     f"cluster:{cluster.name} {cluster.count:2} cluster sum: {cluster.suitenessSum} sum: {sum}\n"
                # )

                for k in range(12):
                    bucket[k] += cluster.suitenessCounts[k]

    allCount = np.sum(bucket)
    if allCount > 1:
        average = sum / allCount
    else:
        average = 0

    outFile.write(
#@        "{} {} suites: average suiteness == {:5.3f}\n".format(
        "{} {} suites: average suiteness== {:5.3f} (power==3.00)\n".format(
            comment, allCount, average
        )
    )
    if mode == 0:
        outFile.write(f"{bucket[11]:6d} suites are  outliers\n")
    outFile.write(f"{bucket[0]:6d} suites have suiteness == 0    \n")
    outFile.write(f"{bucket[1]:6d} suites have suiteness >  0 <.1\n")
    for k in range(2, 10):
        outFile.write(
            "{:6d} suites have suiteness >=.{} <.{}\n".format(bucket[k], k - 1, k)
        )
    outFile.write(f"{bucket[10]:6d} suites have suiteness >=.9    \n")



# ***** kinemage output format *****************************************************

def kinemage1Suite(suite, bin, cluster, notes, distance, suiteness, issue, comment, 
                    pointMaster, pointColor):
    suite.pointMaster = pointMaster
    suite.pointColor = pointColor
    suite.notes = notes


# static text: viewing parameters
janesviews = """
@viewid {d e z}
@zoom 1.00
@zslab 200
@ztran 0
@center 197.500 172.300 178.300
@axischoice 2 3 4
@matrix
0.07196 0.11701 -0.99052 -0.00336 0.99312 0.11707 0.99740 -0.00509 0.07186
@2viewid {zag front}
@2zoom 1.00
@2zslab 200
@2ztran 0
@2center 174.091 194.887 207.768
@2axischoice 4 5 7
@2matrix
0.99508 -0.00018 -0.09905 -0.00135 -0.99993 -0.01172 -0.09904 0.0118 -0.99501
@3viewid {a b g}
@3zoom 1.00
@3zslab 200
@3ztran 0
@3center 175.700 189.600 64.100
@3axischoice 5 6 7
@3matrix
0.99955 0.000101 0.030002 0.0002 0.99995 -0.010012 -0.030001 0.010013 0.9995

"""

# static text: static items in the display
kinemageFrame = """
@group {frame} dominant 
@vectorlist {frame} color= white 
P   0.000   0.000   0.000   5.000   0.000   0.000 
P  35.000   0.000   0.000  40.000   0.000   0.000 
P  80.000   0.000   0.000 160.000   0.000   0.000 
P   0.000   0.000   0.000   0.000   5.000   0.000 
P   0.000  35.000   0.000   0.000  40.000   0.000 
P   0.000  80.000   0.000   0.000 160.000   0.000 
P   0.000   0.000   0.000   0.000   0.000   5.000 
P   0.000   0.000  35.000   0.000   0.000  40.000 
P   0.000   0.000  80.000   0.000   0.000 160.000 
P 200.000   0.000   0.000 280.000   0.000   0.000 
P 320.000   0.000   0.000 360.000   0.000   0.000 
P   0.000 200.000   0.000   0.000 280.000   0.000 
P   0.000 320.000   0.000   0.000 360.000   0.000 
P   0.000   0.000 200.000   0.000   0.000 280.000 
P   0.000   0.000 320.000   0.000   0.000 360.000 
@labellist {XYZ} color= white 
{X}  20.000  -5.000  -5.000 
{X} 380.000  -5.000  -5.000 
{Y}  -5.000  20.000  -5.000 
{Y}  -5.000 380.000  -5.000 
{Z}  -5.000  -5.000  20.000 
{Z}  -5.000  -5.000 380.000 
@labellist {mtp} color= green 
{p}  60.000   0.000   0.000 
{t} 180.000   0.000   0.000 
{m} 300.000   0.000   0.000 
{p}   0.000  60.000   0.000 
{t}   0.000 180.000   0.000 
{m}   0.000 300.000   0.000 
{p}   0.000   0.000  60.000 
{t}   0.000   0.000 180.000 
{m}   0.000   0.000 300.000 

"""

def kinemageFinal(suites, outNote):
    """
    Output the content of a kinemage file
    The 3, 2 order may seem odd, but it is a standard
    """
    kinemageHeader(outNote)
    for deltaMinus in (3, 2): 
      for delta in (3, 2):
          binGroupOut(deltaMinus, delta, suites)
    triaged = bins[0]
    if triaged.count > 0:
        outFile.write("@group {triaged} dominant dimension=9 wrap=360 select off\n")
        binOut(triaged, suites)


def kinemageHeader(outNote):
    """ The invariant portion of a kinemage file """
    outFile.write(f"@text\n {outNote.version}\n {outNote.comment}\n")
    outFile.write("@kinemage 1\n")
    outFile.write("@onewidth\n")
    if args.etatheta or args.thetaeta:  # 070524
        outFile.write(
            "@dimension {theta} {delta-1} {epsilon-1} {zeta-1} {alpha} {beta} {gamma} {delta} {eta}\n"
        )
    else:
        outFile.write(
            "@dimension {chi-1} {delta-1} {epsilon-1} {zeta-1} {alpha} {beta} {gamma} {delta} {chi}\n"
        )
    outFile.write(
        "@dimminmax 0.000 360.000 0.000 360.000 0.000 360.000 0.000 360.000 0.000\
 360.000 0.000 360.000 0.000 360.000 0.000 360.000 0.000 360.000\n"
        )
    if outNote.outliers:
      outFile.write("@pointmaster 'O' {outliers}\n")
    if outNote.wannabes:
      outFile.write("@master {wannabees}\n")
    outFile.write(janesviews)
    outFile.write(kinemageFrame)

    # if(LTepsilon) {outFile.write("@pointmaster 'E' {epsilon bad}\n");}
    # else if(LTdelta || LTdeltam)
    #                       {outFile.write("@pointmaster 'D' {delta bad}\n");}
    # else if(LTzeta || LTalpha || LTbeta || LTgamma)
    #               {outFile.write("@pointmaster 'T' {various bad}\n");}
    # if(Loutlier)  {outFile.write("@pointmaster 'O' {outliers}\n");}


def formatAngles(angles, separator):
    strings = ["{:7.3f}".format(a) for a in angles]
    out = separator.join(strings)
    return out


def binGroupOut(deltaMinus, delta, suites):
    # If any bin in the group has data, generate a group header
    groupCount = 0
    for gamma in ("p", "t", "m"):
      bin = bins[(deltaMinus, delta, gamma)]
      groupCount += bin.count
    if groupCount > 0:        
      outFile.write(
        f"@group {{{deltaMinus}{delta}}} recessiveon dimension=9"
        " wrap=360 select animate off\n")

    # generate the data
    for gamma in ("p", "t", "m"):
        bin = bins[(deltaMinus, delta, gamma)]
        if bin.count > 0:
          binOut(bin, suites)


def binOut(bin, suites):
    if any([c.count > 0 for c in bin.cluster]):
        outFile.write(f"@subgroup {{{bin.name}}} recessiveon \n")
    for cluster in bin.cluster[1:]:
        # the first cluster, for outliers, will be handled later
        if cluster.count > 0:
            extras = ""
            if cluster.status == "wannabe":
                extras = " master= {wannabees}"
            # display a ball for each for each suite in this cluster
            ballList = (
                "@balllist {{{} {}}} color= {} radius= 1 "
                "nohilite master= {{data}}{}\n"
            ).format(bin.name, cluster.name, cluster.clusterColor, extras)
            # display a ring surrounding the center of the cluster
            ringList = (
                "@ringlist {{{} {}}} color= {} radius= 10 width= 1 "
                "nobutton master= {{avsigma}}{}\n"
            ).format(bin.name, cluster.name, cluster.clusterColor, extras)
            angleList = formatAngles(cluster.angle[1:-1], ' ')
            ringList2 = "{{{} {}}} 180 {} 180\n".format(
                bin.name, cluster.name, angleList
            )
            labelList = (
                "@labellist {{{} {}}} color= {} nobutton "
                "master= {{labels}}{}\n"
            ).format(bin.name, cluster.name, cluster.clusterColor, extras)

            outFile.write(ballList)
            outPoints(bin, cluster, suites, "")
            outFile.write(ringList)
            outFile.write(ringList2)
            outFile.write(labelList)
            outFile.write(ringList2)
            
    # handle outliers if there are any:
    if bin.cluster[0].count > 0:
        cluster = bin.cluster[0]
        ballList = (
            "@balllist {{{} {}}} color= {} radius= 1 "
            "nohilite master= {{data}}\n"
        ).format(bin.name, cluster.name, cluster.clusterColor)
        outFile.write(ballList)
        outPoints(bin, cluster, suites, "'O' white")
#@        outPoints(bin, cluster, suites, "")


def outPoints(bin, cluster, suites, extra1):
    for s in suites:
        if s.cluster is cluster:
            if bin.name == "trig":  # the triage bin is especially handled
                extra = f"'{s.pointMaster}'"
            else:
                extra = extra1
            ids = ':'.join(s.pointID)
            line = \
              (f"{{{bin.name} {cluster.name} :D=={s.distance:5.3f}"
               f":S=={s.suiteness:5.3f}: {ids}}} {extra} ,") \
               + formatAngles(s.angle, ",") + "\n"
            outFile.write(line)
   



import suiteninit

import numpy as np
from numpy import array
from enum import Enum


# reasons why a suite may fail to be classified:
Issue = Enum('Issue', 'DELTA_M EPSILON_M ZETA_M ALPHA BETA GAMMA DELTA')
reasons = {
  Issue.DELTA_M:    "delta-1",
  Issue.EPSILON_M:  "epsilon-1",
  Issue.ZETA_M:     "zeta-1",
  Issue.ALPHA:      "alpha",
  Issue.BETA:       "beta",
  Issue.GAMMA:      "gamma",
  Issue.DELTA:      "delta"
}

failMessages = {
  Issue.DELTA_M:    "bad deltam",
  Issue.GAMMA:      "g out",
  Issue.DELTA:      "bad delta"
}


# primary (coarse grained) classification of suites
class Bin:
    # permanence properties
    name = ""
    ordinal = 0
    cluster = ()
    # a tuple of cluster objects
    dominant = -1
    # statistics gathered during the run
    count = 0
    active = False

    def __init__(self, ordinal, name, clusters=()):
        self.ordinal = ordinal
        self.name = name
        self.cluster = clusters
        self.dominant = -1
        self.active = False

        for i, c in enumerate(clusters):
            if c.dominance == "dom":
                self.dominant = i
                break


# secondary (fine grained) classification of suite
class Cluster:
    # intrinsic data:
    ordinal = 0       # its place in the bin
    name = ""         # the name of the cluster
    status = ""       # certain, wannabe, triaged, outlier, nothing, incomplete
    clusterColor = "" # kinemage color names
    dominance = ""    # dom, sat, ord, out, tri, inc
    satelliteInfo = None    # present only if this cluster is a satellite
  
    # tuple of 9 angles: chi-1 as 0 and chi as 8
    # the standard 7 angles are indices 1-7:
    angle = () 

    # gathered statistics:
    count = 0  # number of data points found in this cluster
    suitenessSum = 0
    suitenessCounts = None

    def __init__(self, ordinal, name, status, color, dominance, angles):
        self.ordinal = ordinal
        self.name = name
        self.LOK = (name != "!!")
        self.status = status
        self.clusterColor = color
        self.dominance = dominance  
        self.angle = array(angles)
        if self.dominance == "sat":
            self.satelliteInfo = suiteninit.getSatelliteInfo(name)
        else:
            self.satelliteInfo = None
        self.suitenessCounts = np.zeros(12)
        self.suitenessSum = 0


class SatelliteInfo:
    # numbers used when suite is between satellite and dominant centers
    name = ""
    satelliteWidths = ()   # vector of 9 angles
    dominantWidths = ()    # vector of 9 angles
    
    def __init__(self, name, satelliteWidths, dominantWidths):
        self.name = name
        self.satelliteWidths = satelliteWidths
        self.dominantWidths = dominantWidths


# Residue: a raw residue as normally represented
class Residue:
    '''
    A residue as normally read in.
    Used only briefly as input.
    '''
    pointIDs = []
    base = " "    # A, C, G, U, ...
    # The 6 angles:
    alpha = 0
    beta = 0
    gamma = 0
    delta = 0
    epsilon = 0
    zeta = 0
    chi = 0
    angle = np.full(7, 0.0)

    def __init__(self, ID, base, angles):
        self.pointIDs = ID
        self.base = base
        self.angle = angles
        self.unpackAngles()

    def unpackAngles(self):
        self.alpha = self.angle[0]
        self.beta = self.angle[1]
        self.gamma = self.angle[2]
        self.delta = self.angle[3]
        self.epsilon = self.angle[4]
        self.zeta = self.angle[5]
        # for the future:
        # we can accept chi as a seventh angle if provided
        if len(self.angle) > 6:
            self.chi = self.angle[6]
        else:
            self.chi = -431602080.00  #180
            # A preposterous compromise with the past for now
            #self.chi = -180
    

# Suite: the 
class Suite:
    '''
    The set of angles forming the linkage BETWEEN residues.
    This is the core data structure used in most operations of the program.
    '''    
    pointID = ()
    base = " "    # A, C, G, U, ...
    #chiMinus = 0
    deltaMinus = 0
    epsilon = 0
    zeta = 0
    alpha = 0
    beta = 0
    gamma = 0
    delta = 0
    chi = 0
    # dual representation: individual angles are named for clarity
    # array is for convenience of computation
    angle = np.full(9, 0.0) 

    @property
    def chiMinus(self):
        return self.angle[0]

    @chiMinus.setter
    def chiMinus(self, value):
        self.angle[0] = value

    # fields computed during analysis:
    cluster = None  # The cluster to which it is assigned
    suiteness = 0.0
    distance = 0.0
    notes = ""
    pointMaster = ""
    pointColor = ""

    def __init__(self, ID, base, angles=None):
        self.pointID = ID
        self.base = base
        if angles is not None:
          self.angle = angles
          self.unpackAngles()
        self.cluster = None
        self.suiteness = 0.0
        self.distance = 0.0
        self.notes = ""

    def validate(self):
        # make sure that angles deltaMinus through delta are reasonable
        for i in range(1, 8):
            if self.angle[i] < 0 or self.angle[i] > 360:
                return False
        return True

    def gatherAngles(self):
        self.angle = array((self.chiMinus, self.deltaMinus, self.epsilon, self.zeta, 
            self.alpha, self.beta, self.gamma, self.delta, self.chi))

    def unpackAngles(self):
        self.chiMinus = self.angle[0]
        self.deltaMinus = self.angle[1]
        self.epsilon = self.angle[2]
        self.zeta = self.angle[3]
        self.alpha = self.angle[4]
        self.beta = self.angle[5]
        self.gamma = self.angle[6]
        self.delta = self.angle[7]
        self.chi = self.angle[8]






# LIBTBX_SET_DISPATCHER_NAME phenix.suitename
# LIBTBX_SET_DISPATCHER_NAME molprobity.suitename

import libtbx.load_env
from libtbx.utils import Usage
import os, sys, subprocess

def run(args):
  exe = libtbx.env.under_build("suitename/exe/suitename")
  if (os.name == "nt"):
    exe += ".exe"
  if (not os.path.isfile(exe)):
    from libtbx.str_utils import show_string
    from libtbx.utils import Sorry
    raise Sorry("Missing phenix.suitename executable: %s" % show_string(exe))
  subprocess.call([exe] + args)

if (__name__ == "__main__"):
  run(sys.argv[1:])



# This file is no longer in use, until further notice


# This variable has been experimental but we have settled on 3
power = 3


def hyperEllipsoidDistance(suiteAngles, clusterAngles, nAngles, widthArray):
    if nAngles == 4: workRange = range(2,6)
    else:            workRange = range(1,8)

    summation = 0
    for k in workRange:
        delta = abs(suiteAngles[k] - clusterAngles[k])
        delta = delta/widthArray[k]
        delToPower = pow(delta,power)
        summation = summation+delToPower
    result=pow(summation, 1/power)
    return result


def hyperEllipsoidDistance2(suiteAngles, clusterAngles, nAngles, widthArray):
    if nAngles == 4: workRange = range(2,6)
    else:            workRange = range(1,8)

    del2Power = [pow(abs(suiteAngles[k] - clusterAngles[k])/widthArray[k], power) for k in workRange]
    summation = sum(del2Power)
    result=pow(summation, 1/power)
    return result
       

    
    



from suitenamedefs import Suite, Residue
from suiteninit import args

import numpy as np
import math, sys

altidfield = args.altidfield  # where to find codes for alternatives

# The great variety of codes that may represent each base in the input file
NAListA = ":ADE:  A:A  : Ar:ATP:ADP:AMP:T6A:1MA:RIA:  I:I  :"
NAListG = ":GUA:  G:G  : Gr:GTP:GDP:GMP:GSP:1MG:2MG:M2G:OMG: YG: 7MG:YG :"
NAListC = ":CYT:  C:C  : Cr:CTP:CDP:CMP:5MC:OMC:"
NAListU = ":URA:URI:  U: Ur:U  :UTP:UDP:UMP:5MU:H2U:PSU:4SU:"
NAListT = ":THY:  T:T  : Tr:TTP:TDP:TMP:"
IgnoreDNAList = ": DA: DG: DC: DT:"


# out of the noise, determine the base
def findBase(baseCode):
  if len(baseCode) != 3:
        return 'Z'
    
  if NAListA.find(baseCode) >= 0:   base='A'
  elif NAListG.find(baseCode) >= 0: base='G'
  elif NAListC.find(baseCode) >= 0: base='C'
  elif NAListU.find(baseCode) >= 0: base='U'
  elif NAListT.find(baseCode) >= 0: base='T'
  elif IgnoreDNAList.find(baseCode) >= 0:
    return None  # we ignore DNA residues
  else:  
    base='Y'
  return base


def stringToFloat(string):
 try:
  n = float(string)
 except ValueError:
  n = 9999.0  # or maybe math.nan?
 return n


def readResidues(inFile):
  lines = inFile.readlines()
  residues = []
  for line in lines:
    if len(line.strip()) == 0 or line[0] == '#':  # blank or comment line
      continue
    fields = line.split(':')
    ids = fields[:args.pointidfields]
    print(args.pointidfields, ids, line)
    
    baseCode = fields[args.pointidfields-1]
    angleStrings = fields[args.pointidfields:]
    if ids[altidfield].strip() != "" and ids[altidfield] != args.altidval:
      continue  # lines for the wrong alternative conformation are ignored

    base = findBase(baseCode)
    if not base:    # ignore DNA bases
      continue
    angles = np.array([stringToFloat(s) for s in angleStrings])
    for i in range(len(angles)):
        if angles[i] < 0:
            angles[i] += 360.0

    residue = Residue(ids, base, angles)
    residues.append(residue)
  return residues


def readKinemageFile(inFile):
  """
  We glean the following information from a kinemage file:
  The @dimension command gives us the number of dimensions in the data
  Anything between a @balllist command and a subsequent @ command
  is a data line.
  """  
  lines = inFile.readlines()
  goodLines = []
  place, line = findPrefixInList(lines, "@dimension")
  if place > 0:
      items = line.split()
      dimension = len(items) - 1
  else:
      dimension = args.anglefields
      place = 0
  while place >= 0:
      begin, line = findPrefixesInList(lines, "@balllist", "@dotlist", place)
      if begin > 0:
          end, line = findPrefixInList(lines, "@", begin+1)
          place = end
          if end < 0: end = len(lines)
          goodLines += lines[begin + 1 : end]
      else:
          break
  if len(goodLines) == 0:
      goodLines = lines  # assume a pure data file
  return readKinemageSuites(goodLines, dimension)


def readKinemageSuites(lines, dimension):
  """Read a list of kinemage data lines to yield a suite."""
  suites = []
  for line in lines:
    if len(line.strip()) == 0 or line[0] == '#':  # blank or comment line
      continue
    # A meaningful line begins with an id string enclosed in braces
    if line[0] == '{':
      mark = line.find('}')
      if mark > 0:
        idString = line[1:mark]
        ids = idString.split(':')

        # there may be some miscellaneous markers after the id string
        k = mark + 1
        while k < len(line) and not line[k].isdigit():
          k = k + 1
        mark2 = k

        # once we see a number, everything else is angles
        angleText = line[mark2:]
        angleStrings = angleText.split(' ')
        angleStrings2 = angleText.split(',')
        if len(angleStrings2) > len(angleStrings):
          angleStrings = angleStrings2
        angleList = [stringToFloat(s) for s in angleStrings]
        if len(angleList) != dimension:
          continue  # wrong number of dimensions means probably not a data point
        if dimension == 9:
            angles = np.array(angleList)
        else:  # given only 7 angles,skipping the chi angles on the ends
            angles = np.array([180.0] + angleList + [180.0])
        for i in range(len(angles)):
          if angles[i] < 0:
            angles[i] += 360.0
            
        suite = Suite(ids, 'X', angles)
        suites.append(suite)
  return suites


def findPrefixInList(list, prefix, start=0):
    for i, s in enumerate(list[start:]):
        if s.startswith(prefix):
            return i + start, s
    return -1, None

def findPrefixesInList(list, prefix1, prefix2, start=0):
    for i, s in enumerate(list[start:]):
        if s.startswith(prefix1) or s.startswith(prefix2):
            return i + start, s
    return -1, None


def buildSuiteBetweenResidues(r1, r2):
  suite = Suite(r2.pointIDs, r2.base)
  suite.chiMinus = r1.chi
  suite.deltaMinus = r1.delta
  suite.epsilon = r1.epsilon
  suite.zeta = r1.zeta
  suite.alpha = r2.alpha
  suite.beta = r2.beta
  suite.gamma = r2.gamma
  suite.delta = r2.delta
  suite.chi = r2.chi

  suite.gatherAngles()
  return suite


def buildSuiteFirst(r2):
  suite = Suite(r2.pointIDs, r2.base)
  suite.alpha = r2.alpha
  suite.beta = r2.beta
  suite.gamma = r2.gamma
  suite.delta = r2.delta
  suite.epsilon = 999
  suite.zeta = 999
  suite.chiMinus = 999
  suite.deltaMinus = 999

  suite.gatherAngles()
  return suite


def buildSuiteLast(r1):
  suite = Suite((),"")
  suite.chiMinus = r1.chi
  suite.deltaMinus = r1.delta
  suite.epsilon = r1.epsilon
  suite.zeta = r1.zeta
  suite.alpha = 999
  suite.beta = 999
  suite.gamma = 999
  suite.delta = 999
  suite.chi = 999

  suite.gatherAngles()
  return suite


def buildSuites(residues): 
  suites = [buildSuiteFirst(residues[0])]
  for i in range(len(residues) - 1):
    suites.append(buildSuiteBetweenResidues(residues[i], residues[i+1]))
  suites.append(buildSuiteLast(residues[-1]))
  return suites
